/************************************************************************************
					 _     _           _     _       _     _   
					| |   (_)         | |   (_)     | |   | |  
					| |    ___   _____| |    _  __ _| |__ | |_ 
					| |   | \ \ / / _ \ |   | |/ _` | '_ \| __|
					| |___| |\ V /  __/ |___| | (_| | | | | |_ 
					\_____/_| \_/ \___\_____/_|\__, |_| |_|\__|
												__/ |          
BETA Firmware 								   |___/
							
LiveLight USB Interface with ATMEGA32
Written by Henri Rantanen 

Date		5.6.2011
Version		1.0
Target		Atmel ATMEGA32-16U

************************************************************************************/

#define Channels			27				// Amount of output channels 
#define F_CPU 				14745600		// Clock speed 14,7456 MHz
#define BAUD_RATE 			9600    		// Set Baud-rate
#define PREFIX				0xAA			// Header byte
#define LATENCY				10				// Max color change per input cycle
#define ShutDownDelay		10				// Delay in witch to turn off LED's

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

/* Data input protocol example
	+--------+----+----+----+----+----+----+----+----+----+-\
	| Header | R1 | G1 | B1 | R2 | G2 | B2 | R3 | G3 | B3 |  \
	+--------+----+----+----+----+----+----+----+----+----+---\
*/

// Global Variables
volatile unsigned char rxBuffer[(Channels + 1)];	// Receive buffer	
volatile unsigned char bufferIndex = 0;				// Buffer index value
volatile unsigned char PWM = 0x00; 					// PWM index value
volatile unsigned char channel[(Channels + 1)];		// Channel brightness table
unsigned char NoInputCounter = 0;					// Shut down counter

//Function prototypes
void PWMswitch(void);
void Initialize(void);
void RedStripes(void);
void FadeOut(void);

// Main program
int main(void)
{
	Initialize();
	
	while(1)
	{                                                                                              
	}
}

// UART Receive interrupt
ISR(USART_RXC_vect)	
{
	TCNT0 = 0;									// Reset Shut-Down timer
	
	if (TIMSK & ~0x01)
	{
		TIMSK |= (1 << TOIE0);					// Re-enable TIMER0 interrupt
	}
	
	// Read data from UART
	rxBuffer[bufferIndex] = UDR;				// Read char to buffer
	
	if (rxBuffer[0] == PREFIX)					// Is Prefix as it should?
	{
		if (rxBuffer[bufferIndex] >= channel[bufferIndex])
		{
			// If new value is bigger than old
			if (rxBuffer[bufferIndex] > channel[bufferIndex] + LATENCY)
			{
				channel[bufferIndex] = channel[bufferIndex] + LATENCY;		
			} 
			else
			{
				channel[bufferIndex] = rxBuffer[bufferIndex];
			}
		}
		else
		{
			// If new value is smaller than old
			if (rxBuffer[bufferIndex] < channel[bufferIndex] - LATENCY)
			{
				channel[bufferIndex] = channel[bufferIndex] - LATENCY;
			} 
			else
			{
				channel[bufferIndex] = rxBuffer[bufferIndex];
			}
		}		
	}	
	
	bufferIndex++;										// Grow buffer index
						
	// If buffer is full or prefix is faulty
	if (bufferIndex >= (Channels + 1 ) || rxBuffer[0] != PREFIX) 
	{
		bufferIndex = 0;								// Clear buffer index
	}
}

// TIMER0 Overflow interrupt
ISR(TIMER0_OVF_vect)
{
	// Test if new data is coming
	NoInputCounter++;							// Grow Shut Down counter
	
	// If not, shut LED's down
	if (NoInputCounter > ShutDownDelay)			// If there is no new prefix
	{
		NoInputCounter = 0;
		
		FadeOut();
		RedStripes();
		
		TIMSK = TIMSK & ~(1 << TOIE0);			// Disable TIMER0 interrupt
	} 
}

// TIMER2 Overflow interrupt
ISR(TIMER2_OVF_vect)
{
	// PWM frequency calibration
	//TCNT2 = 166;								// PWM speed for camcorder
	
	PWMswitch();								// Run PWM switch routine
	
	// PWM frequency calibration
	TCNT2 = 250;								// PWM speed for naked eye
	
}

// PWM Switch routine
void PWMswitch(void)
{
// Right Bottom RGB Channel
		if(channel[1] > PWM)  		// 1. channel switch
		{
			PORTA |= 0x04;
		}
		else
		{
			PORTA = PORTA & ~0x04;
		}

		if(channel[2] > PWM) 		// 2. channel switch
		{
			PORTA |= 0x10;
		}
		else
		{
			PORTA = PORTA & ~0x10;
		}

		if(channel[3] > PWM) 		// 3. channel switch
		{
			PORTA |= 0x8;
		}
		else
		{
			PORTA = PORTA & ~0x8;
		}

// Right Top RGB Channel
		if(channel[4] > PWM) 		// 4. channel switch
		{
			PORTB |= 0x04;
		}
		else
		{
			PORTB = PORTB & ~0x04;
		}
		if(channel[5] > PWM) 		// 5. channel switch
		{
			PORTA |= 0x02; 
		}
		else
		{
			PORTA = PORTA & ~0x02;
		}
		
		if(channel[6] > PWM) 		// 6. channel switch
		{
			PORTB |= 0x02; 
		}
		else
		{
			PORTB = PORTB & ~0x02;
		}
	
// Top Right RGB Channel
		if(channel[7] > PWM) 		// 7. channel switch
		{
			PORTA |= 0x01;
		}
		else
		{
			PORTA = PORTA & ~0x01;
		}
		
		if(channel[8] > PWM) 		// 8. channel switch
		{
			PORTB |= 0x08;
		}
		else
		{
			PORTB = PORTB & ~0x08;
		}
		
		if(channel[9] > PWM) 		// 9. channel switch
		{
			PORTB |= 0x01;
		}		
		else
		{
			PORTB = PORTB & ~0x01;
		}


// Top Center RGB Channel
		if(channel[10] > PWM) 		// 10. channel switch
		{
			PORTD |= 0x08;
		}
		else
		{
			PORTD = PORTD & ~0x08;
		}
		
		if(channel[11] > PWM) 		// 11. channel switch
		{
			PORTD |= 0x02;
		}
		else
		{
			PORTD = PORTD & ~0x02;
		}
		
		if(channel[12] > PWM)		// 12. channel switch
		{
			PORTD |= 0x04;
		}
		else
		{
			PORTD = PORTD & ~0x04;
		}

// Top Left RGB Channel
		if(channel[13] > PWM) 		// 13. channel switch
		{
			PORTD |= 0x10;
		}
		else
		{
			PORTD = PORTD & ~0x10;
		}
		
		if(channel[14] > PWM) 		// 14. channel switch
		{
			PORTD |= 0x40;
		}
		else
		{
			PORTD = PORTD & ~0x40;
		}

		if(channel[15] > PWM)		// 15. channel switch
		{
			PORTD |= 0x20;
		}
		else
		{
			PORTD = PORTD & ~0x20;
		}

// Left Top RGB Channel
		if(channel[16] > PWM) 		// 16. channel switch
		{
			PORTC |= 0x01;
		}
		else
		{
			PORTC = PORTC & ~0x01;
		}
		
		if(channel[17] > PWM) 		// 17. channel switch
		{
			PORTD |= 0x80;
		}
		else
		{
			PORTD = PORTD & ~0x80;
		}

		if(channel[18] > PWM)		// 18. channel switch
		{
			PORTC |= 0x02;
		}
		else
		{
			PORTC = PORTC & ~0x02;
		}
// Left Bottom RGB Channel

		if(channel[19] > PWM) 		// 19. channel switch
		{
			PORTC |= 0x10;
		}
		else
		{
			PORTC = PORTC & ~0x10;
		}
		
		if(channel[20] > PWM) 		// 20. channel switch
		{
			PORTC |= 0x04;
		}
		else
		{
			PORTC = PORTC & ~0x04;
		}
		
		if(channel[21] > PWM)		// 21. channel switch
		{
			PORTC |= 0x08;
		}
		else
		{
			PORTC = PORTC & ~0x08;
		}
		
// Bottom Left RGB Channel
		if(channel[22] > PWM) 		// 22. channel switch
		{
			PORTC |= 0x20;
		}
		else
		{
			PORTC = PORTC & ~0x20;
		}
		
		if(channel[23] > PWM) 		// 23. channel switch
		{
			PORTC |= 0x80;
		}
		else
		{
			PORTC = PORTC & ~0x80;
		}
		
		if(channel[24] > PWM)		// 24. channel switch
		{
			PORTC |= 0x40;
		}
		else
		{
			PORTC = PORTC & ~0x40;
		}	
		
//Bottom Right RGB Channel
	

		if(channel[25] > PWM)		// 25. channel switch
		{
			PORTA |= 0x20;
		}
		else
		{
			PORTA = PORTA & ~0x20;
		}

		if(channel[26] > PWM)		// 26. channel switch
		{
			PORTA |= 0x40;
		}
		else
		{
			PORTA = PORTA & ~0x40;
		}

		if(channel[27] > PWM)		// 27. channel switch
		{
			PORTA |= 0x80;
		}
		else
		{
			PORTA = PORTA & ~0x80;
		}

		PWM++;   
}

// Initialize MCU
void Initialize(void)
{
	// Initialize TIMER0
	TCCR0 |= (1 << CS02)|(0 << CS01)|(1 << CS00);	// Timer Prescaler 1/1024
	TIMSK |= (0 << TOIE0);							// Disable Overflow Interrupt
	TCNT0 = 0;										// Initialize Counter
	
	// Initialize TIMER2
	TCCR2 |= (0 << CS22)|(1 << CS21)|(0 << CS20);	// Timer Prescaler 1/8
	TIMSK |= (1 << TOIE2);							// Enable Overflow Interrupt
	TCNT2 = 0;										// Initialize Counter
	
	// Initialize UART
	UBRRH = 0;										// UART Speed HIGH
	UBRRL = ((F_CPU / 16) / BAUD_RATE) - 1; 		// UART Speed LOW
	UCSRB = (1 << 4);                 				// Enable receiver       
	UCSRC = 0x86;									// Character size 8 bits
	UCSRB |= 0x80;									// Enable RX interrupt

	// Initialize output pins
	DDRA = 0xFF;
	DDRB = 0xFF;
	DDRC = 0xFF;
	DDRD = 0xFE;
	
	// Enable global interrupts
	sei();
	
	// Go to red bars
	RedStripes();
}

// Fade in Red bars
void RedStripes(void)
{
	volatile unsigned char i = 0;	// index variable
	
	TIMSK = TIMSK & ~(1 << TOIE2);						// Disable TIMER2 interrupt
	
	// Fade RED sides on
	for (i = 0; i < (128 / (LATENCY / 4)); i++)			// Use latency for fade speed
	{
	
		PWM = 0;										// Start new PWM counter
		while (PWM < 255)								// Run full PWM cycle
		{
			PWMswitch();
		}
	
		channel[1] = channel[1] + (LATENCY / 4);		// Right Bottom
		channel[4] = channel[4] + (LATENCY / 4);		// Right Top
		channel[16] = channel[16] + (LATENCY / 4);		// Left Top
		channel[19] = channel[19] + (LATENCY / 4);		// Left Bottom
	}

	TIMSK |= (1 << TOIE2);								// Enable TIMER2 Interrupt	
	
	bufferIndex = 0;
	rxBuffer[0] = 0;
}

// Fade all LED's off
void FadeOut(void)
{
	volatile unsigned char i = 0;	// index variable
	volatile unsigned char j = 0;	// index variable
	
	TIMSK = TIMSK & ~(1 << TOIE2);						// Disable TIMER2 interrupt
	
	for (i = 0; i < 254; i++)							// Use latency for fade speed
	{
		for (j = 1; j <= Channels; j++)					// all channels -
		{
			if (channel[j] != 0)
			{
				channel[j]--;		// decrease brightness of all channels
			} 				
		}
		
		PWM = 0;										// Start new PWM counter
		while (PWM < 255)								// Run full PWM cycle
		{
			PWMswitch();
		}
	}
	
}